/* Copyright (c) 2008-2017 Holobloc Inc. All rights reserved. */
package fb.rt;

import java.awt.AWTException;
import java.awt.Robot;
import java.io.IOException;

import fb.datatype.ANY;
import fb.datatype.ANYImpl;
import fb.datatype.FBDataTypeException;
import fb.datatype.UINT;
import junit.framework.*;

/**
 * The abstract superclass for testing subclasses of AbstractAdapter.
 * @author JHC
 * @version 20170324/JHC - Removed final qualifier from testConnectParameter_0()
 *          - PRXY_RES does not support connectParameter().
 */
// 20140801/JHC - Added LSEP.
// 20100818/JHC - Added testToString().
// 20100810/JHC - Added testCopyValue(), testEncodeDecode(),
// testInitializeString().
// 20100524/JHC - Moved testConnectOVxxx() methods here from FBInstanceTestCase.
// 20100412/JHC - Moved checkLog(), checkLogSet() logic to EventLogger.
// Removed log.
// 20090427/JHC - Deleted abstract testConnectIV(),
// testConnectIVNoException(), testConnectParameter(), testEINamed(),
// testEONamed(), testIVNamed(), testIVNamedNoException(),
// testOVNamed(), testOVNamedNoException() - subclasses are not
// required to implement these if the corresponding I/O lists are empty.
// 20090425/JHC - Updated testConnectIV_0.
// 20090206/JHC - Added 2-argument static checkLog() to check arbitrary logs.
// 20090112/JHC - Corrected testConnectParameter() for case when invalid object
// is a component of the bad value.
// 20090110/JHC - Added NO_EVENTS.
// 20090109/JHC - Added logger, log, checkLog(), NO_EVENTS.
// 20090105/JHC - Updated testConnectIV(String, ANY) for JUnit 3; added
// ivIdentityTest().
// 20081231/JHC - Adapted from FBRT_CLDC.
public abstract class AbstractAdapterTestCase extends TestCase {
	/** The OS-dependent line separator. */
	public static String LSEP = java.lang.System.getProperty("line.separator");

	private static Robot newRobot() {
		try {
			return new Robot();
		} catch (AWTException e) {
			return null;
		}
	}

	public static final Robot ROBOT = newRobot();

	public static final Robot getRobot() {
		return ROBOT;
	}

	/** An empty array of EventServers for use in testing. */
	public static final EventServer[] NO_EVENTS = new EventServer[0];

	/** Answer an instance of the type being tested. */
	protected abstract AbstractAdapter getFixture();

	/** The sequence of events logger. */
	protected EventLogger logger = new EventLogger();

	/** Checks for correct response to null or invalid IV names. */
	public final void testConnectIV_0() {
		try {
			getFixture().connectIV(null, null);
			fail("Undetected null IV value");
		} catch (FBRManagementException e) {
			assertEquals("INVALID_OBJECT", e.getMessage());
		}
		try {
			getFixture().connectIV("???", new UINT());
			fail("Undetected invalid IV name: \"???\"");
		} catch (FBRManagementException e) {
			assertEquals("NO_SUCH_OBJECT", e.getMessage());
		}
		try {
			getFixture().connectIV(null, new UINT());
			fail("Undetected null IV name");
		} catch (FBRManagementException e) {
			assertEquals("NO_SUCH_OBJECT", e.getMessage());
		}
	}

	/** Checks for correct response to null or invalid IV names. */
	public final void testConnectIVNoException_0() {
		getFixture().connectIVNoException(null, null);
		getFixture().connectIVNoException("???", null);
	}

	/** Checks for correct response to null or invalid parameter names. */
	public void testConnectParameter_0() {
		try {
			getFixture().connectParameter(null, null);
			fail("Undetected null IV name");
		} catch (FBDataTypeException e) {
			fail("Unexpected DataTypeException: " + e.getMessage());
		} catch (FBRManagementException e) {
			assertEquals("NO_SUCH_OBJECT", e.getMessage());
		}
		try {
			getFixture().connectParameter("???", null);
			fail("Undetected invalid IV name: \"???\"");
		} catch (FBRManagementException e) {
			assertEquals("NO_SUCH_OBJECT", e.getMessage());
		} catch (FBDataTypeException e) {
			fail("Unexpected DataTypeException: " + e.getMessage());
		}
	}

	/** Checks for correct response to null or invalid EI names. */
	public final void testEINamed_0() {
		String message = "NO_SUCH_OBJECT";
		try {
			getFixture().eiNamed(null);
			fail(message);
		} catch (FBRManagementException e) {
			assertEquals(message, e.getMessage());
		}
		try {
			getFixture().eiNamed("???");
			fail(message);
		} catch (FBRManagementException e) {
			assertEquals(message, e.getMessage());
		}
	}

	/** Checks for correct response to null or invalid EO names. */
	public void testEONamed_0() {
		String message = "NO_SUCH_OBJECT";
		try {
			getFixture().eoNamed(null);
			fail(message);
		} catch (FBRManagementException e) {
			assertEquals(message, e.getMessage());
		}
		try {
			getFixture().eoNamed("???");
			fail(message);
		} catch (FBRManagementException e) {
			assertEquals(message, e.getMessage());
		}
	}

	/** Checks for correct response to null or invalid IV names. */
	public void testIVNamed_0() {
		try {
			getFixture().ivNamed(null);
			fail("Undetected null IV name");
		} catch (FBRManagementException e) {
			assertEquals("NO_SUCH_OBJECT", e.getMessage());
		}
		try {
			getFixture().ivNamed("???");
			fail("Undetected invalid IV name: \"???\"");
		} catch (FBRManagementException e) {
			assertEquals("NO_SUCH_OBJECT", e.getMessage());
		}
	}

	/** Checks for correct response to null or invalid OV names. */
	public final void testOVNamed_0() {
		try {
			getFixture().ovNamed(null);
			fail("Undetected null OV name");
		} catch (FBRManagementException e) {
			assertEquals("NO_SUCH_OBJECT", e.getMessage());
		}
		try {
			getFixture().ovNamed("???");
			fail("Undetected invalid OV name: \"???\"");
		} catch (FBRManagementException e) {
			assertEquals("NO_SUCH_OBJECT", e.getMessage());
		}
	}

	/** Checks for correct response to null or invalid OV names. */
	public final void testOVNamedNoException_0() {
		assertNull(getFixture().ovNamedNoException(null));
		assertNull(getFixture().ovNamedNoException("???"));
	}

	/**
	 * Checks that <TT>connectIVNamed()</TT> makes a valid connection and
	 * performs proper type checking, including <code>null</code>.
	 * @param ivName The name of the input variable.
	 * @param iv The input variable itself, for comparison.
	 * @param goodvar A variable that should be assignable to the IV without
	 *        error.
	 * @param badvar A variable that should not be assignable to the IV without
	 *        error.
	 * @throws Exception Test requirements not met.
	 */
	protected final void testConnectIV(String ivName, ANY iv, ANY goodvar, ANY badvar) throws Exception {
		// Assign goodvar
		getFixture().connectIV(ivName, goodvar);
		ivIdentityTest(ivName, goodvar);
		// Assignment of badvar should fail
		try {
			getFixture().connectIV(ivName, badvar);
			fail("Uncaught bad IV");
		} catch (FBRManagementException e) {
			assertEquals("INVALID_OBJECT", e.getMessage());
			ivIdentityTest(ivName, goodvar);
		}
		// Assignment of null should fail
		try {
			getFixture().connectIV(ivName, null);
			fail("Uncaught null IV");
		} catch (FBRManagementException e) {
			assertEquals("INVALID_OBJECT", e.getMessage());
			ivIdentityTest(ivName, goodvar);
		}
	}

	/**
	 * Checks that <TT>connectIVNamedNoException()</TT> makes a valid connection
	 * and makes no connection if there is a type conflict.
	 * @param ivName The name of the input variable.
	 * @param iv The input variable itself, for comparison.
	 * @throws Exception Test requirements not met.
	 */
	protected final void testConnectIVNoException(String ivName, ANY iv) throws Exception {
		AbstractAdapter fixture = getFixture();
		String ivType = iv.getType();
		ANY newvar = iv.copy();
		getFixture().connectIVNoException(ivName, newvar);
		assertTrue(fixture.ivNamed(ivName) == newvar);
		if (newvar.isANY())
			return;
		String badType = ivType.equals("INT") ? "UINT" : "INT";
		ANY badvar = ANY.createVar(badType);
		fixture.connectIVNoException(ivName, badvar);
		assertTrue(fixture.ivNamed(ivName) == newvar);
	}

	/**
	 * Checks that <TT>connectParameter()</TT> makes a valid connection and
	 * performs proper type checking.
	 * @param ivName The name of the input variable.
	 * @param initialValue An initial value to set.
	 * @param goodValue The value to set and check.
	 * @param badValue A bad value for exception checking - if this value is
	 *        <code>null</code>, this check is not performed.
	 * @throws Exception Test requirements not met.
	 */
	protected final void testConnectParameter(String ivName, String initialValue, String goodValue, String badValue)
			throws Exception {
		String fbType = getFixture().getType();
		String description = fbType + ".connectParameter(\"" + ivName + "\",\"" + initialValue + "\")";
		ANY goodvar = null;
		ANY iv = null;
		iv = getFixture().ivNamed(ivName);
		getFixture().connectParameter(ivName, initialValue);
		description = fbType + ".connectParameter(\"" + ivName + "\",\"" + goodValue + "\")";
		try {
			goodvar = iv.copy();
			goodvar.initialize(goodValue);
			getFixture().connectParameter(ivName, goodValue);
			if (!iv.equals(goodvar))
				fail(description);
		} catch (FBRManagementException e) {
			fail("FBRManagementException: " + e.getMessage() + " in " + description);
		} catch (FBDataTypeException e) {
			fail("FBDataTypeException: " + e.getMessage() + " in " + description);
		}
		description = fbType + ".connectParameter(\"" + ivName + "\",null)";
		try {
			getFixture().connectParameter(ivName, null);
			fail("Uncaught null value in " + description);
		} catch (FBDataTypeException e) {
			assertEquals("INVALID_OBJECT: null", e.getMessage());
			if (!iv.equals(goodvar))
				fail("Null data changed IV value in " + description);
		} catch (FBRManagementException e) {
			assertEquals("INVALID_OBJECT", e.getMessage());
			if (!iv.equals(goodvar))
				fail("Null data changed IV value in " + description);
		}
		if (badValue == null)
			return;
		description = fbType + ".connectParameter(\"" + ivName + "\",\"" + badValue + "\")";
		try {
			getFixture().connectParameter(ivName, badValue);
			fail("Uncaught bad data in " + description);
		} catch (FBDataTypeException e) {
			assertTrue(e.getMessage().startsWith("INVALID_OBJECT"));
			if (!iv.equals(goodvar))
				fail("Bad data changed IV value in " + description);
		} catch (FBRManagementException e) {
			assertEquals("INVALID_OBJECT", e.getMessage());
			if (!iv.equals(goodvar))
				fail("Bad data changed IV value in " + description);
		}
	}

	/**
	 * Checks that <TT>eiNamed()</TT> returns the correct value for an event
	 * input.
	 * @param eiName The name of the event input.
	 * @param ei The event input itself, for comparison.
	 * @throws Exception Test requirements not met.
	 */
	protected final void testEINamed(String eiName, EventServer ei) throws Exception {
		if (getFixture().eiNamed(eiName) != ei) {
			String fbType = getFixture().getType();
			fail(fbType + ".eiNamed(\"" + eiName + "\") != " + fbType + '.' + eiName);
		}
	}

	/**
	 * Checks that <TT>eoNamed()</TT> returns the correct value for an event
	 * output.
	 * @param eoName The name of the event output.
	 * @param eo The event output itself, for comparison.
	 * @throws AssertionFailedException Test requirements not met.
	 */
	protected final void testEONamed(String eoName, EventServer eo) throws Exception {
		if (getFixture().eoNamed(eoName) != eo) {
			String fbType = getFixture().getType();
			fail(fbType + ".eoNamed(\"" + eoName + "\") != " + fbType + '.' + eoName);
		}
	}

	/**
	 * Test the FB's functions in response to input events.
	 * @throws AssertionFailedException The FB does not function correctly.
	 * @throws DataTypeException An unexpected error occurred.
	 * @throws CommunicationException A communication error was detected.
	 * @throws IOException An error in data transmission or reception.
	 */
	public abstract void testEvents() throws Exception;

	public abstract void testGetType() throws Exception;

	/**
	 * Checks that <TT>ivNamed()</TT> returns the correct value for an input
	 * variable.
	 * @param ivName The name of the input variable.
	 * @param iv The input variable itself, for comparison.
	 * @throws Exception Test requirements not met.
	 */
	protected final void testIVNamed(String ivName, ANY iv) throws Exception {
		String fbType = getFixture().getType();
		try {
			if (getFixture().ivNamed(ivName) != iv) {
				fail(fbType + ".ivNamed(\"" + ivName + "\") != " + fbType + '.' + ivName);
			}
		} catch (FBRManagementException e) {
			fail("FBRManagementException: " + e.getMessage() + " in " + fbType + ".ivNamed(\"" + ivName + "\")");
		}
	}

	/**
	 * Checks that <TT>ovNamed()</TT> returns the correct value for an output
	 * variable.
	 * @param ovName The name of the output variable.
	 * @param ov The output variable itself, for comparison.
	 * @throws Exception Test requirements not met.
	 */
	protected final void testOVNamed(String ovName, ANY ov) throws Exception {
		String fbType = getFixture().getType();
		try {
			if (getFixture().ovNamed(ovName) != ov) {
				fail(fbType + ".ovNamed(\"" + ovName + "\") != " + fbType + '.' + ovName);
			}
		} catch (FBRManagementException e) {
			fail("FBRManagementException: " + e.getMessage() + " in " + fbType + ".ovNamed(\"" + ovName + "\")");
		}
	}

	/**
	 * Checks that <TT>ovNamedNoException()</TT> returns the correct value for
	 * an output variable.
	 * @param ovName The name of the output variable.
	 * @param ov The output variable itself, for comparison.
	 * @throws Exception Test requirements not met.
	 */
	protected final void testOVNamedNoException(String ovName, ANY ov) throws Exception {
		String fbType = getFixture().getType();
		if (getFixture().ovNamedNoException(ovName) != ov) {
			fail(fbType + ".ovNamed(\"" + ovName + "\") != " + fbType + '.' + ovName);
		}
	}

	/**
	 * Tests that the IV with the given name is identical (==) to the given
	 * variable.
	 */
	private final void ivIdentityTest(String ivName, ANY var) throws Exception {
		ANY newvar = getFixture().ivNamed(ivName);
		if (newvar.isANY())
			assertTrue(((ANYImpl) newvar).value == var);
		else
			assertTrue(newvar == var);
	}

	/**
	 * Checks that <TT>connectIVNamed()</TT> makes a valid connection and
	 * performs proper type checking.
	 * @param ivName The name of the input variable.
	 * @param iv The input variable itself, for comparison.
	 * @throws Exception Test requirements not met.
	 */
	protected final void testConnectIV(String ivName, ANY iv) throws Exception {
		AbstractAdapter fixture = getFixture();
		String ivType = iv.getType();
		ANY newvar = iv.copy();
		fixture.connectIV(ivName, newvar);
		assertTrue(fixture.ivNamed(ivName) == newvar);
		try {
			fixture.connectIV(ivName, null);
			fail("Uncaught null IV");
		} catch (FBRManagementException e) {
			assertEquals("INVALID_OBJECT", e.getMessage());
			assertTrue(fixture.ivNamed(ivName) == newvar);
		}
		if (iv.isANY())
			return;
		String badType = ivType.equals("INT") ? "UINT" : "INT";
		try {
			ANY badvar = ANY.createVar(badType);
			getFixture().connectIV(ivName, badvar);
			fail("Uncaught data type conflict");
		} catch (FBRManagementException e) {
			assertEquals("INVALID_OBJECT", e.getMessage());
			assertTrue(fixture.ivNamed(ivName) == newvar);
		}
	}

	/**
	 * Check that the log contents are equal to contents of the given array.
	 * @param sequence The expected event sequence.
	 * @throws Exception The check failed.
	 */
	protected void checkLog(EventServer[] sequence) throws Exception {
		String result = logger.checkLog(sequence);
		if (!result.isEmpty())
			fail(result);
	}

	/**
	 * Check that the log contents are equal to contents of the given array, but
	 * not in any particular order.
	 * @param sequence The expected event sequence.
	 * @throws Exception The check failed.
	 */
	protected void checkLogSet(EventServer[] sequence) throws Exception {
		String result = logger.checkLogSet(sequence);
		if (!result.isEmpty())
			fail(result);
	}

	/**
	 * Checks that <TT>connectOV()</TT> makes a valid connection and performs
	 * proper type checking.
	 * @param ovName The name of the output variable.
	 * @param newOV The new output variable.
	 * @throws Exception Test requirements not met.
	 */
	protected final void testConnectOV(String ovName, ANY newOV) throws Exception {
		AbstractAdapter fixture = getFixture();
		ANYImpl ov = (ANYImpl) fixture.ovNamed(ovName);
		fixture.connectOV(ovName, newOV);
		if (newOV.isANY())
			assertTrue(ov.value == ((ANYImpl) newOV).value);
		else
			assertTrue(ov.value == newOV);
		ANY newOV2 = newOV.copy();
		fixture.connectOV(ovName, newOV2);
		if (newOV2.isANY())
			assertTrue(ov.value == ((ANYImpl) newOV2).value);
		else
			assertTrue(ov.value == newOV2);
		ANY newOV3 = newOV.copy();
		ANYImpl newANY = new ANYImpl(newOV3);
		fixture.connectOV(ovName, newANY);
		assertTrue(ov.value == newOV3);
		newANY.value = null;
		fixture.connectOV(ovName, newANY);
		assertTrue(ov.value == newOV3);
		assertTrue(newANY.value == newOV3);
		try {
			fixture.connectOV(ovName, null);
			fail("Uncaught null IV");
		} catch (FBRManagementException e) {
			assertEquals("INVALID_OBJECT", e.getMessage());
			assertTrue(ov.value == newOV3);
		}
		String ovType = newOV.getType();
		String badType = ovType.equals("INT") ? "UINT" : "INT";
		try {
			ANY badvar = ANY.createVar(badType);
			fixture.connectOV(ovName, badvar);
			fail("Uncaught data type conflict");
		} catch (FBRManagementException e) {
			assertEquals("INVALID_OBJECT", e.getMessage());
			assertTrue(ov.value == newOV3);
		}
	}

	public final void testConnectOV_0() throws Exception {
		AbstractAdapter fbInstance = getFixture();
		try {
			fbInstance.connectOV(null, null);
			fail("Undetected null OV value");
		} catch (FBRManagementException e) {
			assertEquals(FBRManagementException.INVALID_OBJECT, e);
		}
		try {
			fbInstance.connectOV(null, new UINT());
			fail("Undetected null OV name");
		} catch (FBRManagementException e) {
			assertEquals(FBRManagementException.NO_SUCH_OBJECT, e);
		}
		try {
			fbInstance.connectOV("???", new UINT());
			fail("Undetected invalid OV name: \"???\"");
		} catch (FBRManagementException e) {
			assertEquals(FBRManagementException.NO_SUCH_OBJECT, e);
		}
	}

	/**
	 * Checks that <TT>connectOV()</TT> throws
	 * {@link fb.rt.FBRManagementException#INVALID_OPERATION} when connectOV()
	 * is invalid.
	 * @param ovName The name of the output variable.
	 * @param ov The output variable itself.
	 * @throws Exception Test requirements not met.
	 */
	protected final void testConnectOVInvalidOp(String ovName, ANY ov) throws Exception {
		try {
			getFixture().connectOV(ovName, ov.copy());
			fail("INVALID_OPERATION");
		} catch (FBRManagementException e) {
			assertEquals("INVALID_OPERATION", e.getMessage());
		}
	}

	/**
	 * Checks that <TT>connectOVNoException()</TT> makes a valid connection and
	 * makes no connection if there is a type conflict.
	 * @param ovName The name of the output variable.
	 * @param newOV The new output variable.
	 * @throws Exception Test requirements not met.
	 */
	protected final void testConnectOVNoException(String ovName, ANY newOV) throws Exception {
		AbstractAdapter fixture = getFixture();
		String ovType = newOV.getType();
		fixture.connectOVNoException(ovName, newOV);
		ANYImpl ov = (ANYImpl) fixture.ovNamed(ovName);
		if (newOV.isANY()) {
			assertTrue(ov.value == ((ANYImpl) newOV).value);
			return;
		}
		assertTrue(ov.value == newOV);
		String badType = ovType.equals("INT") ? "UINT" : "INT";
		ANY badvar = ANY.createVar(badType);
		fixture.connectOVNoException(ovName, badvar);
		assertTrue(ov.value == newOV);
	}

	/**
	 * Checks that <TT>connectOVNamedNoException()</TT> makes no connection when
	 * that would be an invalid operation.
	 * @param ovName The name of the output variable.
	 * @param ov The output variable itself, for comparison.
	 * @throws Exception Test requirements not met.
	 */
	protected final void testConnectOVNoExceptionInvalidOp(String ovName, ANY ov) throws Exception {
		AbstractAdapter fixture = getFixture();
		ANY newvar = ov.copy();
		fixture.connectOVNoException(ovName, newvar);
		assertTrue(fixture.ovNamed(ovName) == ov);
	}

	/**
	 * Exercises copyValue(), which does nothing for AbstractAdapter and its
	 * subclasses.
	 */
	public void testCopyValue() {
		getFixture().copyValue(null);
	}

	/**
	 * Exercises encodeLV(), decodeLV(), encodeTag() and decodeTag(), which do
	 * nothing for AbstractAdapter and its subclasses.
	 */
	public void testDecode() {
		getFixture().encodeLV(null);
		getFixture().decodeLV(null);
		getFixture().encodeTag(null);
		getFixture().decodeTag(null);
	}

	/**
	 * Exercises initialize(String) and initialize(String, Resource).
	 * @throws Exception
	 */
	public void testInitialize() throws Exception {
		AbstractAdapter fixture = getFixture();
		fixture.initialize("???");
		assertFalse("???".equals(fixture.getName()));
		fixture.initialize("FOO", null);
		assertEquals("FOO", fixture.getName());
		fixture.initialize(null, null);
		assertEquals("FOO", fixture.getName());
	}

	/**
	 * Tests that getName() returns the empty string. May be over-ridden by
	 * tests that actually use the name.
	 * @throws Exception The test failed.
	 */
	public void testGetName() throws Exception {
		assertEquals("", getFixture().getName());
	}

	/**
	 * Tests the <code>toString()</code> method. By default, tests that the
	 * result is in the <code>"Name : Type"</code> format. Subclasses may
	 * over-ride if their <code>toString()</code> method returns some other
	 * value.
	 */
	public void testToString() throws Exception {
		assertEquals(getFixture().getName() + " : " + getFixture().getType(), getFixture().toString());
	}

	public void testConnectTo() {
		getFixture().connectTo(null);
	}
}