/** Copyright (c)2008-2014 Holobloc Inc. All rights reserved. */
package fb.rt;

import java.util.Vector;

/**
 * A utility class to log events to a Vector.
 * 
 * @author JHC
 * @version 20140505/JHC - Replaced sizeError() with waitForSequence(), added
 *          timed check methods.
 */
// 20100520/JHC - Added toString().
// 20100412/JHC - Added checkLog(), checkLogSet(), buildErrorMessage(), size().
// Deleted getLog().
// 20081230/JHC - Copied from FBRT_CLDC project.
public class EventLogger extends AbstractEvent {
	/** The event log. */
	private Vector<EventServer> log = new Vector<EventServer>(8);

	/** Appends the event to the log. */
	@Override
	public void serviceEvent(EventServer e) {
		log.addElement(e);
	}

	/** Constructs a new logger. */
	public EventLogger() {
		super();
	}

	/** Resets (empties) the log. */
	public void reset() {
		log.removeAllElements();
	}

	/**
	 * Checks that the log contents are equal to contents of the given array.
	 * 
	 * @param sequence The expected event sequence.
	 * @return The empty String if the check succeeded; otherwise a non-empty
	 *         explanation for the failure.
	 */
	public String checkLog(EventServer[] sequence) throws Exception {
		return checkLog(sequence, 0);
	}

	private String contentError(EventServer[] sequence) {
		Vector<EventServer> seq = new Vector<EventServer>(sequence.length);
		for (int j = 0; j < sequence.length; j++)
			seq.add(sequence[j]);
		return ("Expecting " + seq + ", saw " + log);

	}

	/**
	 * Returns the size of the contained log.
	 * 
	 * @return The log size.
	 */
	public int size() {
		return log.size();
	}

	/**
	 * Check that the log contents are equal to contents of the given array, but
	 * not in any particular order.
	 * 
	 * @param sequence The expected event sequence.
	 * @return The empty String if the check succeeded; otherwise a non-empty
	 *         explanation for the failure.
	 */
	public String checkLogSet(EventServer[] sequence) {
		return checkLogSet(sequence, 0);
	}

	@Override
	public String toString() {
		return log.toString();
	}

	/**
	 * Waits for the number of log entries to be equal to the specified size.
	 * 
	 * @param sequence The expected event sequence.
	 * @param maxTime The maximum time to wait (value will be set to 25ms if
	 *        less).
	 * @return The empty String if the number of log entries reaches the
	 *         specified size by the specified time; otherwise a non-empty
	 *         explanation for the failure.
	 */
	private synchronized String waitForSequence(EventServer[] sequence,
			int maxTime) {
		int maxT = Math.max(25, maxTime);
		for (int t = 25; t <= maxT; t += 25) {
			if (log.size() == sequence.length)
				return "";
		}
		return "Event log size: expected " + sequence.length + ", saw "
				+ log.size() + " after " + maxT + "ms";
	}

	/**
	 * Checks that the log contents are equal to contents of the given array.
	 * 
	 * @param sequence The expected event sequence.
	 * @param maxTime The maximum time to wait for the sequence.
	 * @return The empty String if the check succeeded; otherwise a non-empty
	 *         explanation for the failure.
	 */
	public String checkLog(EventServer[] sequence, int maxTime)
			throws Exception {
		final String message = waitForSequence(sequence, maxTime);
		if (!message.isEmpty())
			return message;
		for (int i = 0; i < sequence.length; i++) {
			if (sequence[i] != log.elementAt(i))
				return contentError(sequence);
		}
		return "";
	}

	/**
	 * Check that the log contents are equal to contents of the given array, but
	 * not in any particular order.
	 * 
	 * @param sequence The expected event sequence.
	 * @param maxTime The maximum time to wait in ms.
	 * @return The empty String if the check succeeded; otherwise a non-empty
	 *         explanation for the failure.
	 */
	public String checkLogSet(EventServer[] sequence, int maxTime) {
		final String message = waitForSequence(sequence, maxTime);
		if (!message.isEmpty())
			return message;
		for (int i = 0; i < sequence.length; i++) {
			if (!log.contains(sequence[i]))
				return contentError(sequence);
		}
		return "";
	}
}
